/*
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation. See the file COPYING
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * Copyright (C) 2000-2007 Frank van Maarseveen <frankvm@frankvm.com>
 */

#include "fsindex.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <time.h>
#include <pwd.h>
#include <grp.h>

#include "repr.h"

static char *(*do_encode)(const unsigned char *bin, int len);

void repr_init(void)
{
	do_encode = opt_hex ? base16 : base64;
}

char *base16(const unsigned char *bin, int len)
{
	const char *digit = "0123456789abcdef";
	const unsigned char *end;
	char *buf, *p;
	unsigned u;

	end = bin + len;
	buf = xmalloc(len * 2 + 1);
	p = buf;
	while (bin < end) {
		u = *bin++;
		*p++ = digit[u >> 4];
		*p++ = digit[u & 15];
	}
	*p = 0;
	return buf;
}

/*
 * RFC1521 compliant base-64 encoding.
 */
char *base64(const unsigned char *bin, int len)
{
	const char *digit = "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/";
	const unsigned char *end;
	char *buf, *p;
	unsigned bits;
	int nbits;

	end = bin + len;
	buf = xmalloc((len + 2) / 3 * 4 + 1);	/* 4 digits per 3 bytes plus trailing zero */
	p = buf;
	bits = 0;
	nbits = 0;
	while (bin < end) {
		if (nbits < 6) {
			bits = (bits << 8) | *bin++;
			nbits += 8;
		}
		*p++ = digit[(bits >> (nbits - 6)) & 077];
		nbits -= 6;
	}
	/* There are 0, 2, 4 or 6 remaining bits. Shift in zeros for missing input bits */
	if (nbits)
		*p++ = digit[(bits << (6 - nbits)) & 077];
	*p = 0;
	return buf;
}

char *encode(const unsigned char *bin, int len)
{
	return do_encode(bin, len);
}

/*
 * Escape tab, newline and the backslash itself by a backslash and return a
 * newly allocated string when changed. Otherwise return the original argument.
 */
char *escape(const char *s)
{
	const char *in;
	char *buf, *end, *out;
	char c, c2;
	int old, new;
	const int reserved = 2;

	in = s;
	out = buf = end = NULL;
	do {
		c = *in++;
		switch (c) {
		case '\t':
			c2 = 't';
			break;
		case '\n':
			c2 = 'n';
			break;
		case '\\':
			c2 = c;
			break;
		case '\033':		/* avoid ESC games */
			c2 = 'e';
			break;
		default:
			if (!out)
				continue;
			c2 = 0;
			break;
		}
		if (out >= end) {
			old = out - buf;
			if (!old)
				old = in - s - 1;
			new = 2 * old + 10;
			buf = xrealloc(buf, new + reserved);
			end = buf + new;
			if (!out)
				memcpy(buf, s, old);
			out = buf + old;
		}
		*out++ = c;
		if (c2) {
			out[-1] = '\\';
			*out++ = c2;
		}
	} while (c);
	if (out) {
		s = xrealloc(buf, out - buf);
	}
	return (char *)s;
}

/*
 * Return file mode in ls -l style. The returned buffer will be reused on the next call.
 */
const char *mode2str(mode_t mode)
{
	static char buf[11];	/* drwxr-xr-x */
	static mode_t oldmode;

	if (mode == oldmode && *buf)
		return buf;

	oldmode = mode;
	switch (mode & S_IFMT) {
	case S_IFSOCK:	buf[0] = 's';	break;
	case S_IFLNK:	buf[0] = 'l';	break;
	case S_IFREG:	buf[0] = '-';	break;
	case S_IFBLK:	buf[0] = 'b';	break;
	case S_IFDIR:	buf[0] = 'd';	break;
	case S_IFCHR:	buf[0] = 'c';	break;
	case S_IFIFO:	buf[0] = 'p';	break;
#ifdef S_IFDOOR
	case S_IFDOOR:	buf[0] = 'D';	break;		/* Solaris */
#endif
	default:	buf[0] = '?';	break;
	}
	buf[1] = mode &  S_IREAD	? 'r' : '-';
	buf[2] = mode &  S_IWRITE	? 'w' : '-';
	buf[3] = mode &  S_IEXEC	? 'x' : '-';
	buf[4] = mode & (S_IREAD  >> 3)	? 'r' : '-';
	buf[5] = mode & (S_IWRITE >> 3)	? 'w' : '-';
	buf[6] = mode & (S_IEXEC  >> 3)	? 'x' : '-';
	buf[7] = mode & (S_IREAD  >> 6)	? 'r' : '-';
	buf[8] = mode & (S_IWRITE >> 6)	? 'w' : '-';
	buf[9] = mode & (S_IEXEC  >> 6)	? 'x' : '-';
	if (mode & S_ISUID)
		buf[3] = buf[3] == 'x' ? 's' : 'S';
	if (mode & S_ISGID)
		buf[6] = buf[6] == 'x' ? 's' : 'S';
	if (mode & S_ISVTX)
		buf[9] = buf[9] == 'x' ? 't' : 'T';
	return buf;
}

/*
 * Return a string representation of a user id. The returned buffer will be
 * reused on the next call.
 */
static char *uid2str_buf;
const char *uid2str(uid_t uid)
{
	static int size;
	static uid_t olduid;
	struct passwd *pw;

	if (!size) {
		size = 3 * sizeof (uid_t) + 1;		/* assume 3 decimal digits for every byte */
		uid2str_buf = alc(size);
	}
	if (uid == olduid && *uid2str_buf)
		return uid2str_buf;
	olduid = uid;
	if (!opt_numeric) {
		pw = getpwuid(uid);
		if (!pw && (errno == ENOMEM || errno == EIO))
			die("getpwuid: %s\n", strerror(errno));
		if (pw) {
			if (strlen(pw->pw_name) >= size) {
				free(uid2str_buf);
				size = 2 * strlen(pw->pw_name) + 1;
				uid2str_buf = alc(size);
			}
			strcpy(uid2str_buf, pw->pw_name);
			return uid2str_buf;
		}
	}
	sprintf(uid2str_buf, "%lu", (unsigned long)uid);
	return uid2str_buf;
}

/*
 * Return a string representation of a group id. The returned buffer will be
 * reused on the next call.
 */
static char *gid2str_buf;
const char *gid2str(gid_t gid)
{
	static int size;
	static gid_t oldgid;
	struct group *gr;

	if (!size) {
		size = 3 * sizeof (gid_t) + 1;		/* assume 3 decimal digits for every byte */
		gid2str_buf = alc(size);
	}
	if (gid == oldgid && *gid2str_buf)
		return gid2str_buf;
	oldgid = gid;
	if (!opt_numeric) {
		gr = getgrgid(gid);
		if (!gr && (errno == ENOMEM || errno == EIO))
			die("getgrgid: %s\n", strerror(errno));
		if (gr) {
			if (strlen(gr->gr_name) >= size) {
				free(gid2str_buf);
				size = 2 * strlen(gr->gr_name) + 1;
				gid2str_buf = alc(size);
			}
			strcpy(gid2str_buf, gr->gr_name);
			return gid2str_buf;
		}
	}
	sprintf(gid2str_buf, "%lu", (unsigned long)gid);
	return gid2str_buf;
}

/*
 * Return YYYYmmdd HH:MM:SS representation of the given time_t. The returned
 * buffer will be reused on the next call.
 */
const char *time2str(time_t t)
{
	struct tm	*tm;
	static char	buf[32];		/* 18 should do */

	tm = opt_localtime ? localtime(&t) : gmtime(&t);
	strftime(buf, sizeof (buf) - 1, "%" "Y%m%d %H:%M:%S", tm);
	buf[sizeof (buf) - 1] = 0;
	return buf;
}

void repr_cleanup(void)
{
	free(uid2str_buf);
	free(gid2str_buf);
}
